@@ -33,14 +33,33 @@ static const char * AppName = "Gargoyle " VERSION;
 static const char * LaunchingTemplate = "%s/%s";
 static const char * DirSeparator = "/";
 
-#define MaxBuffer 1024
 char dir[MaxBuffer];
 char buf[MaxBuffer];
 char tmp[MaxBuffer];
 
-char filterlist[] = "";
+char *filterlist[] =
+{
+"All Games|*.taf;*.agx;*.d[0-9][0-9];*.acd;*.a3c;*.asl;*.cas;*.ulx;*.hex;*.jacl;*.j2;*.gam;*.t3;*.z?;*.l9;*.sna;*.mag;*.dat;*.saga;*.blb;*.glb;*.zlb;*.blorb;*.gblorb;*.zblorb",
+"Adrift Games (*.taf)|*.taf",
+"AdvSys Games (*.dat)|*.dat",
+"AGT Games (*.agx)|*.agx;*.d[0-9][0-9]",
+"Alan Games (*.acd,*.a3c)|*.acd;*.a3c",
+"Glulx Games (*.ulx)|*.ulx;*.blb;*.blorb;*.glb;*.gblorb",
+"Hugo Games (*.hex)|*.hex",
+"JACL Games (*.jacl,*.j2)|*.jacl;*.j2",
+"Level 9 (*.l9)|*.l9;*.sna",
+"Magnetic Scrolls (*.mag)|*.mag",
+"Quest Games (*.asl,*.cas)|*.asl;*.cas",
+"Scott Adams Grand Adventures (*.saga)|*saga",
+"TADS 2 Games (*.gam)|*.gam;*.t3",
+"TADS 3 Games (*.t3)|*.gam;*.t3",
+"Z-code Games (*.z?)|*.z[0-9];*.zlb;*.zblorb",
+"All Files|*",
+};
 
-void wininit(void)
+const int filtercount = 15;
+
+void winstart(void)
 {
     gtk_init(NULL, NULL);
 }
@@ -72,7 +91,37 @@ int winargs(int argc, char **argv, char *buffer)
     return (argc == 2);
 }
 
-void winopenfile(char *buffer)
+void winfilteradd(GtkFileChooser *dialog, const char *name, const char *pattern)
+{
+    gchar **patterns;
+    gint i;
+
+    GtkFileFilter *filter = gtk_file_filter_new();
+    gtk_file_filter_set_name(filter, name);
+
+    patterns = g_strsplit(pattern, ";", -1);
+    for(i = 0; patterns[i] != NULL; i++)
+        gtk_file_filter_add_pattern(filter, patterns[i]);
+    g_strfreev(patterns);
+
+    gtk_file_chooser_add_filter(dialog, filter);
+}
+
+void winfilterfiles(GtkFileChooser *dialog)
+{
+    gchar **format;
+    gint i;
+
+    for (i = 0; i < filtercount; i++)
+    {
+        format = g_strsplit(filterlist[i], "|", -1);
+        if (format[0] != NULL && format[1] != NULL)
+            winfilteradd(dialog, format[0], format[1]);
+        g_strfreev(format);
+    }
+}
+
+void winbrowsefile(char *buffer)
 {
     *buffer = 0;
 
@@ -85,6 +134,8 @@ void winopenfile(char *buffer)
     if (getenv("HOME"))
         gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(openDlg), getenv("HOME"));
 
+    winfilterfiles(GTK_FILE_CHOOSER(openDlg));
+
     gint result = gtk_dialog_run(GTK_DIALOG(openDlg));
 
     if (result == GTK_RESPONSE_ACCEPT)
@@ -101,7 +152,8 @@ void winpath(char *buffer)
 
     exelen = readlink("/proc/self/exe", exepath, sizeof(exepath));
 
-    if (exelen <= 0 || exelen >= MaxBuffer) {
+    if (exelen <= 0 || exelen >= MaxBuffer)
+    {
         winmsg("Unable to locate executable path");
         exit(EXIT_FAILURE);
     }
@@ -112,7 +164,7 @@ void winpath(char *buffer)
     if ( dirpos != NULL )
         *dirpos = '\0';
 
-   return;
+    return;
 }
 
 int winexec(const char *cmd, char **args)
@@ -124,7 +176,9 @@ int winterp(char *path, char *exe, char *flags, char *game)
 {
     sprintf(tmp, LaunchingTemplate, path, exe);
 
-    char *args[] = {NULL, NULL, NULL};
+    setenv("GARGLK_INI", path, FALSE);
+
+    char *args[4] = {NULL, NULL, NULL, NULL};
 
     if (strstr(flags, "-"))
     {
@@ -138,7 +192,8 @@ int winterp(char *path, char *exe, char *flags, char *game)
         args[1] = buf;
     }
 
-    if (!winexec(tmp, args)) {
+    if (!winexec(tmp, args))
+    {
         winmsg("Could not start 'terp.\nSorry.");
         return FALSE;
     }
@@ -148,14 +203,14 @@ int winterp(char *path, char *exe, char *flags, char *game)
 
 int main(int argc, char **argv)
 {
-    wininit();
+    winstart();
 
     /* get dir of executable */
     winpath(dir);
 
     /* get story file */
     if (!winargs(argc, argv, buf))
-        winopenfile(buf);
+        winbrowsefile(buf);
 
     if (!strlen(buf))
         return TRUE;
