@@ -58,12 +58,20 @@ static HCURSOR idc_arrow, idc_hand, idc_ibeam;
 
 static MMRESULT timer = 0;
 static int timerid = -1;
-static int timeouts = 0;
+static volatile int timeouts = 0;
 
 /* buffer for clipboard text */
 static wchar_t *cliptext = NULL;
 static int cliplen = 0;
 
+/* filters for file dialogs */
+static char *winfilters[] =
+{
+    "Saved game files (*.sav)\0*.sav\0All files (*.*)\0*.*\0\0",
+    "Text files (*.txt)\0*.txt\0All files (*.*)\0*.*\0\0",
+    "All files (*.*)\0*.*\0\0",
+};
+
 void glk_request_timer_events(glui32 millisecs)
 {
 	if (timerid != -1)
@@ -73,7 +81,7 @@ void glk_request_timer_events(glui32 millisecs)
 		timerid = -1;
 	}
 
-    if (millisecs)
+	if (millisecs)
 	{
 		timeBeginPeriod(1);
 		timer = timeSetEvent(millisecs, 0, timeproc, 0, TIME_PERIODIC);
@@ -139,7 +147,7 @@ void winexit(void)
     exit(0);
 }
 
-void winopenfile(char *prompt, char *buf, int len, char *filter)
+void winopenfile(char *prompt, char *buf, int len, int filter)
 {
     OPENFILENAME ofn;
     memset(&ofn, 0, sizeof(OPENFILENAME));
@@ -149,7 +157,7 @@ void winopenfile(char *prompt, char *buf, int len, char *filter)
     ofn.nMaxFile = len;
     ofn.lpstrInitialDir = NULL;
     ofn.lpstrTitle = prompt;
-    ofn.lpstrFilter = filter;
+    ofn.lpstrFilter = winfilters[filter];
     ofn.nFilterIndex = 1;
     ofn.lpstrDefExt = ("");
     ofn.Flags = OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;
@@ -158,7 +166,7 @@ void winopenfile(char *prompt, char *buf, int len, char *filter)
     strcpy(buf, "");
 }
 
-void winsavefile(char *prompt, char *buf, int len, char *filter)
+void winsavefile(char *prompt, char *buf, int len, int filter)
 {
     OPENFILENAME ofn;
     memset(&ofn, 0, sizeof(OPENFILENAME));
@@ -168,7 +176,7 @@ void winsavefile(char *prompt, char *buf, int len, char *filter)
     ofn.nMaxFile = len;
     ofn.lpstrInitialDir = NULL;
     ofn.lpstrTitle = prompt;
-    ofn.lpstrFilter = filter;
+    ofn.lpstrFilter = winfilters[filter];
     ofn.nFilterIndex = 1;
     ofn.lpstrDefExt = ("");
     ofn.Flags = OFN_OVERWRITEPROMPT;
@@ -405,34 +413,6 @@ void winrepaint(int x0, int y0, int x1, int y1)
     InvalidateRect(hwndview, &wr, 1); // 0);
 }
 
-void winloop(void)
-{
-    MSG msg;
-    int i;
-
-    i = GetMessage(&msg, NULL, 0, 0);
-    if (i < 0)
-    exit(1);
-    if (i > 0)
-    {
-    TranslateMessage(&msg);
-    DispatchMessage(&msg);
-    }
-}
-
-void winpoll(void)
-{
-    MSG msg;
-    int i;
-
-    i = PeekMessage(&msg, NULL, 0, 0, PM_REMOVE);
-    if (i)
-    {
-    TranslateMessage(&msg);
-    DispatchMessage(&msg);
-    }
-}
-
 void gli_select(event_t *event, int polled)
 {
     MSG msg;
@@ -440,6 +420,13 @@ void gli_select(event_t *event, int polled)
     gli_curevent = event;
     gli_event_clearevent(event);
 
+    while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) > 0)
+    {
+        TranslateMessage(&msg);
+        DispatchMessage(&msg);
+    }
+    gli_dispatch_event(gli_curevent, polled);
+
     if (!polled)
     {
         while (gli_curevent->type == evtype_None && !timeouts)
@@ -456,16 +443,6 @@ void gli_select(event_t *event, int polled)
         }
     }
 
-    else
-    {
-        while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) > 0 && !timeouts)
-        {
-            TranslateMessage(&msg);
-            DispatchMessage(&msg);
-        }
-        gli_dispatch_event(gli_curevent, polled);
-    }
-
     if (gli_curevent->type == evtype_None && timeouts)
     {
         gli_event_store(evtype_Timer, NULL, 0, 0);
@@ -497,7 +474,7 @@ void winresize(void)
 
 void CALLBACK timeproc(UINT uID, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2)
 {
-	PostMessage(hwndframe, WM_TIMER, 0, 0);
+    PostMessage(hwndframe, WM_TIMER, 0, 0);
 }
 
 LRESULT CALLBACK
