@@ -1,6 +1,7 @@
 /******************************************************************************
  *                                                                            *
  * Copyright (C) 2006-2009 by Tor Andersson.                                  *
+ * Copyright (C) 2010 by Ben Cressey, Chris Spiegel.                          *
  *                                                                            *
  * This file is part of Gargoyle.                                             *
  *                                                                            *
@@ -30,10 +31,9 @@
 static zchar statusline[256];
 static int oldstyle = 0;
 static int curstyle = 0;
-static int upperstyle = 0;
-static int lowerstyle = 0;
 static int cury = 1;
 static int curx = 1;
+static int fixforced = 0;
 
 static int curr_fg = -2;
 static int curr_bg = -2;
@@ -165,10 +165,12 @@ void reset_status_ht(void)
 	{
 		glk_window_get_size(gos_upper, NULL, &height);
 		if (mach_status_ht != height)
+		{
 			glk_window_set_arrangement(
 				glk_window_get_parent(gos_upper),
 				winmethod_Above | winmethod_Fixed,
 				mach_status_ht, NULL);
+		}
 	}
 }
 
@@ -178,6 +180,11 @@ void erase_window (zword w)
 		glk_window_clear(gos_lower);
 	else if (gos_upper)
 	{
+#ifdef GARGLK
+			garglk_set_reversevideo_stream(
+				glk_window_get_stream(gos_upper),
+				TRUE);
+#endif /* GARGLK */
 		memset(statusline, ' ', sizeof statusline);
 		glk_window_clear(gos_upper);
 		reset_status_ht();
@@ -256,8 +263,6 @@ void smartstatusline (void)
 	int roomlen, scorelen, scoreofs;
 	int len, tmp;
 
-	statusline[curx - 1] = 0; /* terminate! */
-
 	packspaces(statusline, packed);
 	//strcpy(packed, statusline);
 	len = os_string_length(packed);
@@ -300,7 +305,6 @@ void smartstatusline (void)
 	//	buf[roomlen + 1] = '|';
 
 	glk_window_move_cursor(gos_upper, 0, 0);
-	glk_set_style(style_Preformatted);
 	glk_put_buffer_uni(buf, h_screen_cols);
 	glk_window_move_cursor(gos_upper, cury - 1, curx - 1);
 }
@@ -320,16 +324,19 @@ void screen_char (zchar c)
 	}
 
 	/* check fixed flag in header, game can change it at whim */
-	if (gos_curwin == gos_lower)
+	int forcefix = ((h_flags & FIXED_FONT_FLAG) != 0);
+	int curfix = ((curstyle & FIXED_WIDTH_STYLE) != 0);
+	if (forcefix && !curfix)
 	{
-		static int forcefix = -1;
-		int curfix = h_flags & FIXED_FONT_FLAG;
-		if (forcefix != curfix)
-		{
-			forcefix = curfix;
-			zargs[0] = 0xf000;	/* tickle tickle! */
-			z_set_text_style();
-		}
+		zargs[0] = 0xf000;	/* tickle tickle! */
+		z_set_text_style();
+		fixforced = TRUE;
+	}
+	else if (!forcefix && fixforced)
+	{
+		zargs[0] = 0xf000;	/* tickle tickle! */
+		z_set_text_style();
+		fixforced = FALSE;
 	}
 
 	if (gos_upper && gos_curwin == gos_upper)
@@ -342,9 +349,12 @@ void screen_char (zchar c)
 		else {
 			if (cury == 1)
 			{
-				if (curx < sizeof statusline)
+				if (curx <= ((sizeof statusline / sizeof(zchar)) - 1))
+				{
 					statusline[curx - 1] = c;
-				curx++;
+					statusline[curx] = 0;
+				}
+				curx ++;
 				if (curx <= h_screen_cols)
 					glk_put_char_uni(c);
 				else
@@ -804,29 +814,25 @@ void z_set_window (void)
 {
 	int win = zargs[0];
 
-	if (gos_curwin == gos_lower)
-		lowerstyle = curstyle;
-	else
-		upperstyle = curstyle;
-
 	if (win == 0)
 	{
 		glk_set_window(gos_lower);
 		gos_curwin = gos_lower;
-		curstyle = lowerstyle;
 	}
 	else
 	{
 		if (gos_upper)
 			glk_set_window(gos_upper);
 		gos_curwin = gos_upper;
-		curstyle = upperstyle;
 	}
 
-		if (win == 0)
-			enable_scripting = TRUE;
-		else
-			enable_scripting = FALSE;
+	if (win == 0)
+		enable_scripting = TRUE;
+	else
+		enable_scripting = FALSE;
+
+	zargs[0] = 0xf000;	/* tickle tickle! */
+	z_set_text_style();
 }
 
 /*
@@ -879,7 +885,6 @@ void z_show_status (void)
 
 	curx = cury = 1;
 	glk_window_move_cursor(gos_upper, 0, 0);
-	glk_set_style(style_Preformatted);
 
 	/* If the screen width is below 55 characters then we have to use
 	   the brief status line format */
